import warnings
warnings.simplefilter("ignore",category=FutureWarning)

import logging
import numpy as np
from timeit import default_timer as timer
import copy
import sys
import os
import codecs
from sklearn import metrics as sklearn_metrics
from training_wrapper.general_params import load_parameters as load_general_parameters
from collections import Counter 


# Use 'config_file' command line parameter for changing the name of the config file used
cf = 'config_dataset'
for arg in sys.argv[1:]:
    k, v = arg.split('=')
    if k == 'config_file':
        cf = v
cf = __import__(cf)
all_models_params = cf.load_parameters()  # parameters for the current dataset version
params = load_general_parameters()  # general parameters

# Load keras, mkw and training_wrapper modules.
sys.path.insert(0, params["MKW_PATH"])
sys.path.insert(0, params["KERAS_PATH"])

from keras_wrapper.cnn_model import saveModel, loadModel
from keras_wrapper.extra import evaluation
from keras_wrapper.extra.callbacks import EvalPerformance  
from keras_wrapper.extra.read_write import *
from keras_wrapper.utils import decode_multilabel, decode_predictions_words_as_classes

from training_wrapper.prepare_data import build_dataset
from training_wrapper.model import Food_Model


logging.basicConfig(level=logging.DEBUG, format='[%(asctime)s] %(message)s', datefmt='%d/%m/%Y %H:%M:%S')
logger = logging.getLogger(__name__)


# Merges model-specific params to general params
def mergeParams(params, specific_params):
    for k,v in specific_params.items():
        params[k] = v
    return params


def is_bb_overlap(boxA, boxB):
    isOverlap = True

    minXA, maxXA = boxA[0] - boxA[2] * .5, boxA[0] + boxA[2] * .5
    minYA, maxYA = boxA[1] - boxA[3] * .5, boxA[1] + boxA[3] * .5

    minXB, maxXB = boxB[0] - boxB[2] * .5, boxB[0] + boxB[2] * .5
    minYB, maxYB = boxB[1] - boxB[3] * .5, boxB[1] + boxB[3] * .5

    if (minXB > maxXA or maxXB < minXA or minYB > maxYA or maxYB < minYA):
        isOverlap = False

    return isOverlap


def intersection_over_minbbox(boxA, boxB):
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[0] - boxA[2] * .5, boxB[0] - boxB[2] * .5)
    yA = max(boxA[1] - boxA[3] * .5, boxB[1] - boxB[3] * .5)
    xB = min(boxA[0] + boxA[2] * .5, boxB[0] + boxB[2] * .5)
    yB = min(boxA[1] + boxA[3] * .5, boxB[1] + boxB[3] * .5)

    # compute the area of intersection rectangle
    interArea = abs(xB - xA) * abs(yB - yA)

    # compute the area of both the prediction and ground-truth
    # rectangles
    boxAArea = boxA[2] * boxA[3]
    boxBArea = boxB[2] * boxB[3]

    boxArea = np.minimum(boxAArea, boxBArea)
    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    intersection = interArea / float(boxArea)

    # return the intersection over union value
    return intersection


def evaluate_model(params, specific_params):
    """
        Function for using a previously trained model for sampling.
    """

    ########### Load data
    dataset, inputs_mapping, outputs_mapping, class_weights = build_dataset(params, specific_params, splits=specific_params['SPLITS'])
    ###########

    params = mergeParams(params, specific_params)

    
    dic_outputs = {}
    num_outputs = len(specific_params['OUTPUTS_IDS_DATASET'])
    for i, id_name in enumerate(specific_params['OUTPUTS_IDS_DATASET']):
        data = specific_params['OUTPUTS'][id_name]
        dic_outputs[id_name] = {}
        dic_outputs[id_name]['f1_scores'] = []

        ########### Fist best epoch for each output
        model_path = params['STORE_PATH']
        metrics_file_path = model_path + "/test." + data['metrics'][0]
        with open(metrics_file_path, 'r') as m_f:
            lines = m_f.readlines()[1:]
            m_f.close()
        for j, line in enumerate(lines):
            if j % num_outputs == i:
                dic_outputs[id_name]['f1_scores'].append(line.split(',')[5])
        dic_outputs[id_name]['reload'] = np.argmax(dic_outputs[id_name]['f1_scores'])+1
        ###########


        ########### Load model
        model = loadModel(model_path, dic_outputs[id_name]['reload'])
        
        # Define the inputs and outputs mapping from our Dataset instance to our model
        model.setInputsMapping(inputs_mapping)
        model.setOutputsMapping(outputs_mapping)
        model.params = params
        model.setOptimizer()
        ###########

        extra_vars = dict()
        extra_vars['n_parallel_loaders'] = params['PARALLEL_LOADERS']

        if data['type'] != 'sigma':
            extra_vars[i] = dict()

        # Prepare GT for each output
        for s in params['EVAL_ON_SETS']:
            if data['type'] == 'categorical':
                num_classes = 0
                with open(specific_params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as f:
                    for line in f: num_classes += 1
                extra_vars[i]['n_classes'] = num_classes
                extra_vars[i][s] = dict()
                extra_vars[i][s]['references'] = eval('dataset.Y_' + s + '["' + id_name + '"]')
            elif data['type'] == 'text':
                extra_vars[i][s] = dict()
                extra_vars[i][s]['references'] = eval('dataset.Y_' + s + '["' + id_name + '"]')
                extra_vars[i][s]['words2idx'] = dataset.vocabulary[id_name]['words2idx']
                extra_vars[i]['words_as_classes'] = data['words_as_classes']
        
            if data['save_classes_report']:
                extra_vars[i]['save_classes_report'] = data['save_classes_report']
                with codecs.open(specific_params['DATA_ROOT_PATH']+'/'+data['classes'], 'r', 'utf-8') as f:
                    lines = f.readlines()
                classes_names = [l.strip('\n') for l in lines]
                num_classes = len(classes_names)
                extra_vars[i]['n_classes'] = num_classes
                extra_vars[i]['target_names'] = classes_names
                extra_vars[i]['save_path'] = model_path

        if data['type'] == 'categorical':
            vocab = ''
        elif data['type'] == 'text':
            vocab = dataset.vocabulary[id_name]['idx2words']

        gt_id_to_eval = id_name
        gt_pos_to_eval = i
        vocab_x = None
        output_type = data['type']
        metrics = data['metrics']
        write_type = data['write_type']
        min_pred_val = data.get('min_pred_val', 0.5)


        ########### Apply sampling

        # Apply model predictions
        params_prediction = {#'max_batch_size': params['BATCH_SIZE'],
                                'n_parallel_loaders': params['PARALLEL_LOADERS'],
                                'predict_on_sets': ['test']}

        predictions = model.predictNet(dataset, params_prediction)['test']

        # In case of multiple outputs, we predict from the last one
        if isinstance(predictions, list) and len(predictions) > 0:
            predictions = predictions[i]

        if extra_vars[i].get('words_as_classes', False):
            predictions = decode_predictions_words_as_classes(predictions,
                                                              1,   # always set temperature to 1
                                                              vocab,
                                                              'max_likelihood',
                                                              verbose=1)

        with open(model_path + "/test_model_predictions_output" + str(i) + ".txt", 'w') as p_f:
            for pred in predictions:
                pred_sentence = ""
                for p in pred:
                    if p == '<pad>':
                        break
                    else:
                        pred_sentence += p + ' '
                p_f.write(pred_sentence[:-1] + '\n')
            p_f.close()


def generate_predictions(params, specific_params):
    """
        Arguments:
         - params: general parameters
         - specific_params: parameters of the specific model

        Return:
         - Dictionary with the following structure:
            - 'preds' = dictionary with numpy array of predictions for each output. Keys are numbers of outputs (0, 1, ...)
            - 'probs' = dictionary with numpy array of probabilities for each prediction of each output. Keys are numbers of outputs (0, 1, ...)
            - 'words2idx' = vocabulary for matching words to index of probabilities
    """

    ########### Load data
    dataset, inputs_mapping, outputs_mapping, class_weights = build_dataset(params, specific_params, splits=specific_params['SPLITS'])
    ###########

    pred_set = params['EVAL_ON_SETS'][0]

    params = mergeParams(params, specific_params)
    
    dic_outputs = {}
    num_outputs = len(specific_params['OUTPUTS_IDS_DATASET'])
    output_preds = []

    for i, id_name in enumerate(specific_params['OUTPUTS_IDS_DATASET']):
        data = specific_params['OUTPUTS'][id_name]
        dic_outputs[id_name] = {}
        dic_outputs[id_name]['f1_scores'] = []

        ########### Find best epoch for each output
        model_path = params['STORE_PATH']
        metrics_file_path = model_path + '/test.' + data['metrics'][0]
        with open(metrics_file_path, 'r') as m_f:
            lines = m_f.readlines()[1:]
            m_f.close()
        for j, line in enumerate(lines):
            if j % num_outputs == i:
                dic_outputs[id_name]['f1_scores'].append(line.split(',')[5])
        dic_outputs[id_name]['reload'] = np.argmax(dic_outputs[id_name]['f1_scores'])+1
        ###########


        ########### Load model
        model = loadModel(model_path, dic_outputs[id_name]['reload'])
        
        # Define the inputs and outputs mapping from our Dataset instance to our model
        model.setInputsMapping(inputs_mapping)
        model.setOutputsMapping(outputs_mapping)
        model.params = params
        model.setOptimizer()
        ###########

        extra_vars = dict()
        extra_vars['n_parallel_loaders'] = params['PARALLEL_LOADERS']

        if data['type'] != 'sigma':
            extra_vars[i] = dict()

        # Prepare GT for each output
        for s in params['EVAL_ON_SETS']:
            if data['type'] == 'categorical':
                num_classes = 0
                with open(specific_params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as f:
                    for line in f: num_classes += 1
                extra_vars[i]['n_classes'] = num_classes
                extra_vars[i][s] = dict()
                extra_vars[i][s]['references'] = eval('dataset.Y_' + s + '["' + id_name + '"]')
            elif data['type'] == 'text':
                extra_vars[i][s] = dict()
                extra_vars[i][s]['references'] = eval('dataset.Y_' + s + '["' + id_name + '"]')
                extra_vars[i][s]['words2idx'] = dataset.vocabulary[id_name]['words2idx']
                extra_vars[i]['words_as_classes'] = data['words_as_classes']
        
            if data['save_classes_report']:
                extra_vars[i]['save_classes_report'] = data['save_classes_report']
                with codecs.open(specific_params['DATA_ROOT_PATH']+'/'+data['classes'], 'r', 'utf-8') as f:
                    lines = f.readlines()
                classes_names = [l.strip('\n') for l in lines]
                num_classes = len(classes_names)
                extra_vars[i]['n_classes'] = num_classes
                extra_vars[i]['target_names'] = classes_names
                extra_vars[i]['save_path'] = model_path

        if data['type'] == 'categorical':
            vocab = ''
        elif data['type'] == 'text':
            vocab = dataset.vocabulary[id_name]['idx2words']

        gt_id_to_eval = id_name
        gt_pos_to_eval = i
        vocab_re = dataset.vocabulary[id_name]['words2idx']
        output_type = data['type']
        metrics = data['metrics']
        write_type = data['write_type']
        min_pred_val = data.get('min_pred_val', 0.5)


        ########### Apply sampling

        # Apply model predictions
        params_prediction = {#'max_batch_size': params['BATCH_SIZE'],
                                'n_parallel_loaders': params['PARALLEL_LOADERS'],
                                'predict_on_sets': params['EVAL_ON_SETS']}

        predictions = model.predictNet(dataset, params_prediction)[pred_set]

        # In case of multiple outputs, we predict from the last one
        if isinstance(predictions, list) and len(predictions) > 0:
            output_preds.append(predictions[i])

    probs_arr = np.asarray(predictions)
    
    dic_preds_probs = {} 
    dic_preds_probs['preds'] = {}
    dic_preds_probs['probs'] = {}
    dic_preds_probs['words2idx'] = vocab_re
    
    if num_outputs == 1:
        dic_preds_probs['probs'][0] = probs_arr
        dic_preds_probs['preds'][0] = decode_predictions_words_as_classes(probs_arr,
                                                        1,   # always set temperature to 1
                                                        vocab,
                                                        'max_likelihood',
                                                        verbose=1)
    else:
        for i in range(num_outputs):
            dic_preds_probs['probs'][i] = probs_arr[i]
            dic_preds_probs['preds'][i] = decode_predictions_words_as_classes(probs_arr[i],
                                                            1,   # always set temperature to 1
                                                            vocab,
                                                            'max_likelihood',
                                                            verbose=1)
    
    return dic_preds_probs


def combine_predictions_double_output_model(dic_preds_probs, params, specific_params):
    """
        Arguments:
         - dic_preds_probs: Dictionary with the following structure:
            - 'preds' = dictionary with numpy array of predictions for each output. Keys are numbers of outputs (0, 1, ...)
            - 'probs' = dictionary with numpy array of probabilities for each prediction of each output. Keys are numbers of outputs (0, 1, ...)
            - 'words2idx' = vocabulary for matching words to index of probabilities
         - params: general parameters
         - specific_params: parameters of the specific model

        Return:
         - Dictionary with the following structure:
            - 'preds' = numpy array of final predictions. Keys are indexes of the images.
            - 'probs' = numpy array of probabilities for each prediction. Keys are indexes of the images.
    """
    
    num_images = dic_preds_probs['probs'][0].shape[0]
    words2idx = dic_preds_probs['words2idx']
    model_path = params['STORE_PATH']

    pred_set = params['EVAL_ON_SETS'][0].upper()

    # Read images names and folds
    with open(specific_params[pred_set+'_IMGS_LIST'] % specific_params['DATA_ROOT_PATH'], "r") as f:
        imgs_content = f.readlines()
        imgs_all = [c.strip('\n') for c in imgs_content]
        f.close()

    img_name_all = []
    img_fold_all = []
    for i, image in enumerate(imgs_all):
        img_name = image.split("/")[-1].split(".")[0]
        if img_name.split("_")[1] != '0':
            continue
        img_name_all.append(img_name.split("_")[0])
        img_fold_all.append(image.split("/")[-2])

    with open(specific_params['DATA_ROOT_PATH'] + "/annotations/f1-score_all_classes_yolo.txt", 'r') as y_f:  #all or test
        content = y_f.readlines()
        yolo_f1classes = [c.strip('\n') for c in content]
        y_f.close()

    # Create empty results dictionary
    res_dic = {}
    res_dic['preds'] = {}
    res_dic['probs'] = {}

    counter = {}
    counter['total'] = 0
    counter['yolo'] = 0
    counter['ensemble'] = 0
    
    for i, img_name in enumerate(img_name_all): # Num of test images

        # Read yolo preds and probs for each image
        img_yolo_pred = "/tray_images_dataset/new_data_test" + img_fold_all[i] + "/prediction/preds_post_iou/" + img_name + ".txt"
        p_f = open(img_yolo_pred, "r")
        p_list = [line.strip('\n').split() for line in p_f]
        yolo_preds = [l[0] for l in p_list]
        yolo_bboxes = [[float(l[1]), float(l[2]), float(l[3]), float(l[4])] for l in p_list]
        yolo_probs = [float(l[5]) for l in p_list]
        p_f.close()

        if params['YOLO_ENSEMBLE']:
            # Read yolo preds and probs for each image (predictions generated with CNMS disabled. Thus, with multiple predictions for the same bounding box)
            img_yolo_pred_rep = "/tray_images_dataset/new_data_test" + img_fold_all[i] + "/prediction/preds_9000_test_multi/" + img_name + ".txt"
            p_f = open(img_yolo_pred_rep, "r")
            p_list = [line.strip('\n').split() for line in p_f]
            yolo_rep = {}
            for l in p_list:
                bbox = l[1] + ' ' + l[2] + ' ' + l[3] + ' ' + l[4]
                if bbox not in yolo_rep:
                    yolo_rep[bbox] = {}
                    yolo_rep[bbox]['preds'] = []
                    yolo_rep[bbox]['probs'] = []
                yolo_rep[bbox]['preds'].append(l[0])
                yolo_rep[bbox]['probs'].append(float(l[5]))
            p_f.close()


        pred_sentence = ""
        res_dic['preds'][i] = []
        res_dic['probs'][i] = []

        for j in range(len(yolo_preds)):
            
            v0 = dic_preds_probs['preds'][0][i][j]                      # Prediction for output 0
            v0_prob = dic_preds_probs['probs'][0][i][j][words2idx[v0]]  # Probability of prediction for output 0
            v1 = dic_preds_probs['preds'][1][i][j]                      # Prediction for output 1
            v1_prob = dic_preds_probs['probs'][1][i][j][words2idx[v1]]  # Probability of prediction for output 1

            # If one of the two preds is '<pad>', we take the other. Otherwise, the take the prediction with max probability
            if v0 == '<pad>' and v1 != '<pad>': 
                p = v1
                p_prob = v1_prob
            elif v0 != '<pad>' and v1 == '<pad>': 
                p = v0
                p_prob = v0_prob
            else:
                if v0_prob > v1_prob:
                    p = v0
                    p_prob = v0_prob
                else:
                    p = v1
                    p_prob = v1_prob


            # Due to the vocabulary doesn't contain label 0 for background (menu vocabulary limitations), we set '<unk>' to 0 
            if p == '<unk>': 
                p = '0'
            
            # If YOLO_ENSEMBLE is True, an ensemble of the predictions from model and YOLO is done
            if params['YOLO_ENSEMBLE']:
                yolo_pred = yolo_preds[j]
                yolo_prob = yolo_probs[j]
                yolo_bbox = yolo_bboxes[j]

                # Search the other secundary preds for the yolo_bbox
                yolo_bbox_key = ""
                for key in yolo_rep:
                    bbox_key = [float(x) for x in key.split()]
                    if intersection_over_minbbox(yolo_bbox, bbox_key) > 0.9:
                        yolo_bbox_key = key
                        break
                
                # Normalizes yolo probs
                if yolo_bbox_key != "":
                    probs_normalized = np.asarray(yolo_rep[yolo_bbox_key]['probs'])             

                if p == '<pad>' or v0 == yolo_pred or v1 == yolo_pred:  

                    # If the best pred from model is '<pad>', then we take the yolo pred
                    if p == '<pad>':  
                        p = yolo_pred
                        p_prob = yolo_prob

                    # If one of the output predictions is the same as the yolo prediction, then we take the yolo pred
                    if v0 == yolo_pred:
                        p = yolo_pred
                        p_prob = max(yolo_prob, v0_prob)
                    elif v1 == yolo_pred:
                        p = yolo_pred
                        p_prob = max(yolo_prob, v1_prob)

                    counter['yolo'] += 1

                else:
                    model_yolo_pred = max(dic_preds_probs['probs'][0][i][j][words2idx[yolo_pred]], dic_preds_probs['probs'][1][i][j][words2idx[yolo_pred]])

                    if p_prob < yolo_prob + model_yolo_pred:

                        p = yolo_pred
                        p_prob = max(yolo_prob, model_yolo_pred)
                        counter['ensemble'] += 1
            
            pred_sentence += p + ' '
            res_dic['preds'][i].append(p)
            res_dic['probs'][i].append(p_prob)

        counter['total'] += j

    print "Num yolo preds: " + str(counter['yolo'])
    print "Num ensembled preds: " + str(counter['ensemble'])
    print "Num total preds: " + str(counter['total'])

    return res_dic


def decode_predictions_single_output_model(dic_preds_probs, params, specific_params):
    """
        Arguments:
         - dic_preds_probs: Dictionary with the following structure:
            - 'preds' = dictionary with numpy array of predictions for each output. Keys are numbers of outputs (0, 1, ...)
            - 'probs' = dictionary with numpy array of probabilities for each prediction of each output. Keys are numbers of outputs (0, 1, ...)
            - 'words2idx' = vocabulary for matching words to index of probabilities
         - params: general parameters
         - specific_params: parameters of the specific model

        Return:
         - Dictionary with the following structure:
            - 'preds' = numpy array of final predictions. Keys are indexes of the images.
            - 'probs' = numpy array of probabilities for each prediction. Keys are indexes of the images.
    """
    

    num_images = dic_preds_probs['probs'][0].shape[0]
    words2idx = dic_preds_probs['words2idx']
    model_path = params['STORE_PATH']

    pred_set = params['EVAL_ON_SETS'][0].upper()

    # Read images names and folds
    with open(specific_params[pred_set+'_IMGS_LIST'] % specific_params['DATA_ROOT_PATH'], "r") as f:
        imgs_content = f.readlines()
        imgs_all = [c.strip('\n') for c in imgs_content]
        f.close()

    img_name_all = []
    img_fold_all = []
    for i, image in enumerate(imgs_all):
        img_name = image.split("/")[-1].split(".")[0]
        if img_name.split("_")[1] != '0':
            continue
        img_name_all.append(img_name.split("_")[0])
        img_fold_all.append(image.split("/")[-2])

    # Create empty results dictionary
    res_dic = {}
    res_dic['preds'] = {}
    res_dic['probs'] = {}

    
    for i, img_name in enumerate(img_name_all): # Num of test images

        # Read yolo preds and probs for each image
        img_yolo_pred = "/tray_images_dataset/new_data_test" + img_fold_all[i] + "/prediction/preds_post_iou/" + img_name + ".txt"
        p_f = open(img_yolo_pred, "r")
        p_list = [line.strip('\n').split() for line in p_f]
        yolo_preds = [l[0] for l in p_list]
        yolo_bboxes = [[float(l[1]), float(l[2]), float(l[3]), float(l[4])] for l in p_list]
        yolo_probs = [float(l[5]) for l in p_list]
        p_f.close()

        if params['YOLO_ENSEMBLE']:
            # Read yolo preds and probs for each image (predictions generated with CNMS disabled. Thus, with multiple predictions for the same bounding box)
            img_yolo_pred_rep = "/tray_images_dataset/new_data_test" + img_fold_all[i] + "/prediction/preds_9000_test_multi/" + img_name + ".txt"
            p_f = open(img_yolo_pred_rep, "r")
            p_list = [line.strip('\n').split() for line in p_f]
            yolo_rep = {}
            for l in p_list:
                bbox = l[1] + ' ' + l[2] + ' ' + l[3] + ' ' + l[4]
                if bbox not in yolo_rep:
                    yolo_rep[bbox] = {}
                    yolo_rep[bbox]['preds'] = []
                    yolo_rep[bbox]['probs'] = []
                yolo_rep[bbox]['preds'].append(l[0])
                yolo_rep[bbox]['probs'].append(float(l[5]))
            p_f.close()


        pred_sentence = ""
        res_dic['preds'][i] = []
        res_dic['probs'][i] = []

        for j in range(len(yolo_preds)):
            
            p = dic_preds_probs['preds'][0][i][j]                      # Prediction for output 0
            p_prob = dic_preds_probs['probs'][0][i][j][words2idx[p]]  # Probability of prediction for output 0

            # Due to the vocabulary doesn't contain label 0 for background (menu vocabulary limitations), we set '<unk>' to 0 
            if p == '<unk>': 
                p = '0'
            
            # If YOLO_ENSEMBLE is True, an ensemble of the predictions from model and YOLO is done
            if params['YOLO_ENSEMBLE']:
                yolo_pred = yolo_preds[j]
                yolo_prob = yolo_probs[j]
                yolo_bbox = yolo_bboxes[j]

                # Search the other secundary preds for the yolo_bbox
                yolo_bbox_key = ""
                for key in yolo_rep:
                    bbox_key = [float(x) for x in key.split()]
                    if intersection_over_minbbox(yolo_bbox, bbox_key) > 0.9:
                        yolo_bbox_key = key
                        break
                
                # Normalizes yolo probs
                if yolo_bbox_key != "":
                    probs_normalized = np.asarray(yolo_rep[yolo_bbox_key]['probs']) / np.sum(yolo_rep[yolo_bbox_key]['probs'])
                    yolo_prob = np.max(probs_normalized)  # Best yolo pred normalized
                

                if p == '<pad>' or p == yolo_pred:  

                    # If the best pred from model is '<pad>', then we take the yolo pred
                    if p == '<pad>':  
                        p = yolo_pred
                        p_prob = yolo_prob       
                    else:
                        p = yolo_pred
                        p_prob = max(yolo_prob, p_prob)

                else:
                    model_yolo_pred = dic_preds_probs['probs'][0][i][j][words2idx[yolo_pred]]

                    if yolo_bbox_key != "" and p in yolo_rep[yolo_bbox_key]['preds']:
                        idx = yolo_rep[yolo_bbox_key]['preds'].index(p)
                        yolo_p_prob = probs_normalized[idx]
                    else:
                        yolo_p_prob = 0
                    
                    if p_prob + yolo_p_prob < yolo_prob + model_yolo_pred:
                        p = yolo_pred
                        p_prob = yolo_prob
            
            pred_sentence += p + ' '
            res_dic['preds'][i].append(p)
            res_dic['probs'][i].append(p_prob)

    return res_dic


def read_classes(all_classes_names_path, yolo_names):

    # Read all classes names and create classes dictionary    
    with open(all_classes_names_path, 'r') as cn:
        classes_list = cn.readlines()

    dic_classes = {}
    for i in range(len(classes_list)):
        dic_classes[classes_list[i].strip('\n')] = i

    # Add mappings to dic_classes (there are different yolo classes that actually are the same)
    dic_classes['cuscus'] = dic_classes['couscous']
    dic_classes['Canalones'] = dic_classes['canelones']
    dic_classes['canalones'] = dic_classes['canelones']
    dic_classes['ensalada de lentejas '] = dic_classes['ensalada de lentejas']
    dic_classes['fideu\xc3\xa1 '] = dic_classes['fideu\xc3\xa1']
    dic_classes['arroz bolonesa'] = dic_classes['arroz a la bolo\xc3\xb1esa']
    dic_classes['arroz con bolonesa'] = dic_classes['arroz a la bolo\xc3\xb1esa']
    dic_classes['Escalopa de ternera'] = dic_classes['escalopa de ternera']
    dic_classes['escalopa de ternera '] = dic_classes['escalopa de ternera']
    dic_classes['Bistec de ternera a la pimienta'] = dic_classes['bistec de ternera a la pimienta']
    dic_classes['ensalada de cuscus'] = dic_classes['ensalada de couscous']
    dic_classes['espinacas a la crema '] = dic_classes['espinacas a la crema']
    dic_classes['empanadillas '] = dic_classes['empanadillas']
    dic_classes['espaguetis napolitana'] = dic_classes['espaguetis']
    dic_classes['tallarines al pesto'] = dic_classes['espaguetis']

    with open(yolo_names, 'r') as g:
        classes_list = g.readlines()
        yolo_classes_list = [c.strip('\n') for c in classes_list]
        g.close()

    return dic_classes, yolo_classes_list


def apply_nms_and_evaluate_metrics(dic_preds_probs, params, specific_params):
    """
        Arguments:
         - dic_preds_probs: Dictionary with the following structure:
            - 'preds' = numpy array of final predictions. Keys are indexes of the images.
            - 'probs' = numpy array of probabilities for each prediction. Keys are indexes of the images.
         - params: general parameters
         - specific_params: parameters of the specific model

        Returns:
         - dic_results: Dictionary with the following structure:
            - 'f1_classes': list of classes F1-score
            - 'f1': Average of the classes F1-scores
            - 'w_f1': Weighted average (by the support for each class) of the classes F1-scores

        Applies a Non Maximum Supression with stricter NMS and CNMS thresholds. After, evaluates the final predictions (background discarded).
    """

    ########### NMS
    # Parameters
    nms_thresh = params['NMS_THRESH']
    cnms_thresh = params['CNMS_THRESH']

    pred_set = params['EVAL_ON_SETS'][0].upper()

    # Read yolo classes
    dic_classes, yolo_classes_list = read_classes(specific_params['ALL_CLASSES_NAMES_PATH'] % specific_params['DATA_ROOT_PATH'], 
                                                  specific_params['YOLO_NAMES'] % specific_params['DATA_ROOT_PATH'])

    # Read gt labels of the set to evaluate
    with open(specific_params[pred_set+'_GT_LABELS'] % specific_params['DATA_ROOT_PATH'], "r") as g:
        gt_content = g.readlines()
        gt_all = [c.strip('\n') for c in gt_content]
        g.close()

    # Read imgs list of the set to evaluate
    with open(specific_params[pred_set+'_IMGS_LIST'] % specific_params['DATA_ROOT_PATH'], "r") as f:
        imgs_content = f.readlines()
        imgs_all = [c.strip('\n') for c in imgs_content]
        f.close()

    # Create lists of images names and folds
    img_name_all = []
    img_fold_all = []
    for i, image in enumerate(imgs_all):
        img_name = image.split("/")[-1].split(".")[0]
        if img_name.split("_")[1] != '0':
            continue
        img_name_all.append(img_name.split("_")[0])
        img_fold_all.append(image.split("/")[-2])

    # Read and create Counter of unique (without background and repetitions) gt labels of the set to evaluate
    gt_counters_list = []
    with open(specific_params[pred_set+'_GT_UNIQUE'] % specific_params['DATA_ROOT_PATH'], "r") as g:
        gt_content = g.readlines()
        for line in gt_content:
            gt_counter = Counter()
            for c in line.strip('\n').split():
                gt_counter[int(c)] += 1
            gt_counters_list.append(gt_counter)
        g.close()


    res_pred = []
    res_gt = []

    for i, img_name in enumerate(img_name_all):
        img_gt = [int(a) for a in gt_all[i].split()]
        img_pred = []
        img_prob = []
        for j, a in enumerate(dic_preds_probs['preds'][i]):
            try:
                a = int(a)
            except:
                if a == "<unk>":
                    a = 0
                else:
                    print "Found element " + str(a) + ". Image index: " + str(i)
                continue
            img_pred.append(a)
            img_prob.append(float(dic_preds_probs['probs'][i][j]))


        img_yolo_pred = "/tray_images_dataset/new_data_test" + img_fold_all[i] + "/prediction/preds_post_iou/" + img_name + ".txt"
        p_f = open(img_yolo_pred, "r")
        p_list = [line.strip() for line in p_f]
        p_f.close()

        lbl_pred_list = []
        lbl_gt_list = []
        bbs_pred = []
        prob_pred = []

        for j, pred in enumerate(p_list):
            if j >= len(img_pred):
                break
            if img_pred[j] == 0:
                continue
            lbl, x, y, bw, bh, prob = pred.split(" ")
            lbl_pred_list.append(img_pred[j])
            lbl_gt_list.append(img_gt[j])
            boxA = [float(x), float(y), float(bw), float(bh)]
            bbs_pred.append(boxA)
            prob_pred.append(float(img_prob[j]))

        aux_lbl_pred_list = lbl_pred_list
        aux_bbs_pred = bbs_pred
        aux_prob_pred = prob_pred
        aux_lbl_gt_list = lbl_gt_list

        sort1 = np.argsort(aux_prob_pred)[::-1]
        aux_prob_pred = list(np.array(aux_prob_pred)[sort1])
        aux_lbl_pred_list = list(np.array(aux_lbl_pred_list)[sort1])
        aux_bbs_pred = list(np.array(aux_bbs_pred)[sort1])
        aux_lbl_gt_list = list(np.array(aux_lbl_gt_list)[sort1])

        selected_bbox = []
        selected_label = []
        init_index = 0

        for k in range(len(aux_lbl_pred_list) - 1):
            if aux_lbl_pred_list[k] == None:
                continue
            selected_bbox.append(aux_bbs_pred[k])
            selected_label.append(int(aux_lbl_pred_list[k]))
            init_index = k + 1
            break

        for k in range(init_index, len(aux_lbl_pred_list)):
            if aux_lbl_pred_list[k] == None:
                continue
            boxB = aux_bbs_pred[k]
            lblB = int(aux_lbl_pred_list[k])
            flag = False
            for j in range(len(selected_bbox)):
                boxA = selected_bbox[j]
                lblA = int(selected_label[j])
                intersection_minibbox = intersection_over_minbbox(boxA, boxB)
                if is_bb_overlap(boxA, boxB) and ((intersection_minibbox > cnms_thresh and lblA != lblB) or (intersection_minibbox > nms_thresh and lblA == lblB)):
                    aux_lbl_pred_list[k] = None
                    aux_bbs_pred[k] = None
                    aux_prob_pred[k] = None
                    aux_lbl_gt_list[k] = None
                    flag = True
                    break
            if not flag:
                selected_bbox.append(boxB)
                selected_label.append(lblB)

        predictions_counter = Counter()
        for j in range(len(aux_lbl_pred_list)):
            if aux_lbl_pred_list[j] != None:
                predictions_counter[aux_lbl_pred_list[j]] += 1


        counters_intersection = gt_counters_list[i] & predictions_counter
        for a in counters_intersection.elements():
            res_gt.append(int(a))
            res_pred.append(int(a))
        
        gt_diff = gt_counters_list[i] - counters_intersection
        pred_diff = predictions_counter - counters_intersection

        max_len = max(sum(gt_diff.values()), sum(pred_diff.values()))
        gt_diff_elems = list(gt_diff.elements()) + ['0']*(max_len - sum(gt_diff.values()))
        pred_diff_elems = list(pred_diff.elements()) + ['0']*(max_len - sum(pred_diff.values()))
        
        for a, b in zip(gt_diff_elems, pred_diff_elems):
            res_gt.append(int(a))
            res_pred.append(int(b))


    # Metrics
    y_gt = np.asarray(res_gt)
    y_pred = np.asarray(res_pred)


    # Compute Precision, Recall and F1 score
    precision, recall, f1_classes, support = sklearn_metrics.precision_recall_fscore_support(y_gt, y_pred, average=None)

    w_recall = np.dot(recall[1:], support[1:]) / np.sum(support[1:])
    w_precision = np.dot(precision[1:], support[1:]) / np.sum(support[1:])
    w_f1 = np.dot(f1_classes[1:], support[1:]) / np.sum(support[1:])

    with open(params['STORE_PATH'] + "/f1-score_classes_output.txt", 'w') as f:
        for v in f1_classes:
            f.write(str(v)+'\n')
        f.close()

    # Don't use background for metrics computation
    support = support[1:]
    precision = precision[1:]
    recall = recall[1:]

    # Precision and recall arrays, only of elements with non zero support
    precision_ = np.nan_to_num(np.take(precision, np.nonzero(support)))
    recall_ = np.nan_to_num(np.take(recall, np.nonzero(support)))

    # F1-score per class
    f1_ = 2*precision_*recall_ / (precision_+recall_)
    f1_ = np.nan_to_num(f1_)  # It is possible to divide by 0 and get nan values. So we set those to 0.0

    # Get mean metrics between all classes
    precision = np.mean(precision_)
    recall = np.mean(recall_)
    f1 = np.mean(f1_)

    with codecs.open(specific_params['ALL_CLASSES_NAMES_PATH'] % specific_params['DATA_ROOT_PATH'], 'r', 'utf-8') as f:
    	lines = f.readlines()
    	classes_names = [l.strip('\n') for l in lines]
     	f.close()

    report = sklearn_metrics.classification_report(y_gt, y_pred, labels=[i for i in range(len(classes_names))], target_names=classes_names)

    with codecs.open(params['STORE_PATH'] + "/classification_report.txt", 'w', 'utf-8') as f:
    	f.write(report)
    	f.close()

    print
    print "OUTPUT METRICS: "
    print
    print "Mean of per class Precision = " + str(precision)
    print "Mean of per class Recall = " + str(recall)
    print "Mean of per class F1-scores = " + str(f1)
    print 
    print "Support-weighted mean of Precision = " + str(w_precision)
    print "Support-weighted mean of Recall = " + str(w_recall)
    print "Support-weighted mean of F1-scores = " + str(w_f1)        
    print
    print "-----------------------------------------------------"
    print  

    dic_results = {}
    dic_results['f1_classes'] = f1_classes
    dic_results['f1'] = f1
    dic_results['w_f1'] = w_f1

    print "Parameters used: "
    print " -> NMS = " + str(nms_thresh)
    print " -> CNMS = " + str(cnms_thresh)
    print

    return dic_results


def apply_nms_and_predict(dic_preds_probs, params, specific_params):
    """
        Arguments:
         - dic_preds_probs: Dictionary with the following structure:
            - 'preds' = numpy array of final predictions. Keys are indexes of the images.
            - 'probs' = numpy array of probabilities for each prediction. Keys are indexes of the images.
         - params: general parameters
         - specific_params: parameters of the specific model

        Applies a Non Maximum Supression with stricter NMS and CNMS thresholds. After, writes the final predictions (background discarded) to specific_params['DATA_ROOT_PATH']/params['RESULTS_FOLDER_NAME'].
    """

    ########### NMS
    # Parameters
    nms_thresh = params['NMS_THRESH']
    cnms_thresh = params['CNMS_THRESH']

    results_folder_name = params['RESULTS_FOLDER_NAME']
    results_path = specific_params['DATA_ROOT_PATH'] + "/" + results_folder_name

    pred_set = params['EVAL_ON_SETS'][0].upper()

    # Read yolo classes
    dic_classes, yolo_classes_list = read_classes(specific_params['ALL_CLASSES_NAMES_PATH'] % specific_params['DATA_ROOT_PATH'], 
                                                  specific_params['YOLO_NAMES'] % specific_params['DATA_ROOT_PATH'])


    # Read imgs list of the set to evaluate
    with open(specific_params[pred_set+'_IMGS_LIST'] % specific_params['DATA_ROOT_PATH'], "r") as f:
        imgs_content = f.readlines()
        imgs_all = [c.strip('\n') for c in imgs_content]
        f.close()

    # Create lists of images names and folds
    img_name_all = []
    img_fold_all = []
    for i, image in enumerate(imgs_all):
        img_name = image.split("/")[-1].split(".")[0]
        if img_name.split("_")[1] != '0':
            continue
        img_name_all.append(img_name.split("_")[0])
        img_fold_all.append(image.split("/")[-2])


    for i, img_name in enumerate(img_name_all):
        img_pred = []
        img_prob = []
        for j, a in enumerate(dic_preds_probs['preds'][i]):
            try:
                a = int(a)
            except:
                if a == "<unk>":
                    a = 0
                else:
                    print "Found element " + str(a) + ". Image index: " + str(i)
                continue
            img_pred.append(a)
            img_prob.append(float(dic_preds_probs['probs'][i][j]))

        img_yolo_pred = "/tray_images_dataset/new_data_test" + img_fold_all[i] + "/prediction/preds_post_iou/" + img_name + ".txt"
        p_f = open(img_yolo_pred, "r")
        p_list = [line.strip() for line in p_f]
        p_f.close()

        lbl_pred_list = []
        bbs_pred = []
        prob_pred = []
        bbs_index = []

        for j, pred in enumerate(p_list):
            if j >= len(img_pred):
                break
            if img_pred[j] == 0:  # Background discarded
                continue
            lbl, x, y, bw, bh, prob = pred.split(" ")
            lbl_pred_list.append(img_pred[j])
            boxA = [float(x), float(y), float(bw), float(bh)]
            bbs_pred.append(boxA)
            bbs_index.append(j)
            prob_pred.append(float(img_prob[j]))

        aux_lbl_pred_list = lbl_pred_list
        aux_bbs_pred = bbs_pred
        aux_bbs_index = bbs_index
        aux_prob_pred = prob_pred

        sort1 = np.argsort(aux_prob_pred)[::-1]
        aux_prob_pred = list(np.array(aux_prob_pred)[sort1])
        aux_lbl_pred_list = list(np.array(aux_lbl_pred_list)[sort1])
        aux_bbs_pred = list(np.array(aux_bbs_pred)[sort1])
        aux_bbs_index = list(np.array(aux_bbs_index)[sort1])

        selected_bbox = []
        selected_label = []
        init_index = 0

        for k in range(len(aux_lbl_pred_list) - 1):
            if aux_lbl_pred_list[k] == None:
                continue
            selected_bbox.append(aux_bbs_pred[k])
            selected_label.append(int(aux_lbl_pred_list[k]))
            init_index = k + 1
            break

        for k in range(init_index, len(aux_lbl_pred_list)):
            if aux_lbl_pred_list[k] == None:
                continue
            boxB = aux_bbs_pred[k]
            lblB = int(aux_lbl_pred_list[k])
            flag = False
            for j in range(len(selected_bbox)):
                boxA = selected_bbox[j]
                lblA = int(selected_label[j])
                intersection_minibbox = intersection_over_minbbox(boxA, boxB)
                if is_bb_overlap(boxA, boxB) and ((intersection_minibbox > cnms_thresh and lblA != lblB) or (intersection_minibbox > nms_thresh and lblA == lblB)):
                    aux_lbl_pred_list[k] = None
                    aux_bbs_pred[k] = None
                    aux_prob_pred[k] = None
                    aux_bbs_index[k] = None
                    flag = True
                    break
            if not flag:
                selected_bbox.append(boxB)
                selected_label.append(lblB)

        lbl_pred_list = []
        bbs_pred = []
        bbs_index = []
        prob_pred = []
        for j in range(len(aux_lbl_pred_list)):
            if aux_lbl_pred_list[j] != None:
                lbl_pred_list.append(aux_lbl_pred_list[j])
                bbs_pred.append(aux_bbs_pred[j])
                prob_pred.append(aux_prob_pred[j])
                bbs_index.append(aux_bbs_index[j])

        res = []
        for k in range(len(prob_pred)):
            lbl = lbl_pred_list[k]
            bbs = bbs_pred[k]
            prob = prob_pred[k]
            index = bbs_index[k]
            str_dets = " ".join([str(lbl), str(bbs[0]), str(bbs[1]), str(bbs[2]), str(bbs[3]), str(prob), str(index)]) + "\n"
            res.append(str_dets)
        with open(results_path + "/" + img_name + ".txt", "w") as f2:
            f2.writelines(res)
