import warnings
warnings.simplefilter("ignore",category=FutureWarning)

import logging
import numpy as np
from timeit import default_timer as timer
import copy
import sys
import os
import codecs
import time

from training_wrapper.general_params import load_parameters as load_general_parameters

from evaluate_models import evaluate_model, generate_predictions, combine_predictions_double_output_model, decode_predictions_single_output_model, apply_nms_and_evaluate_metrics, apply_nms_and_predict

logging.basicConfig(level=logging.DEBUG, format='[%(asctime)s] %(message)s', datefmt='%d/%m/%Y %H:%M:%S')
logger = logging.getLogger(__name__)


def train_model(params, specific_params):
    """
        Training function
    """

    ########### Load data
    if 'SPLITS' in specific_params.keys():
        dataset, inputs_mapping, outputs_mapping, class_weights = build_dataset(params, specific_params, empty_label=specific_params['EMPTY_LABEL'], splits=specific_params['SPLITS']) # dataset used for evaluation
    else:
        dataset, inputs_mapping, outputs_mapping, class_weights = build_dataset(params, specific_params, empty_label=specific_params['EMPTY_LABEL'])
    ###########

    ########### Build model
    params = mergeParams(params, specific_params)
    if params['RELOAD'] == 0:
        food_model = Food_Model(params,
                            type=specific_params['MODEL_TYPE'], verbose=params['VERBOSE'],
                            empty_label=params['EMPTY_LABEL'],
                            model_name=specific_params['MODEL_NAME'],
                            store_path=specific_params['STORE_PATH'])
    else:
        # Reload model previusly trained
        model_path = params['STORE_PATH'] % params['MODELS_ROOT_PATH']    
        food_model = loadModel(model_path, params['RELOAD'])

    # Define the inputs and outputs mapping from our Dataset instance to our model
    food_model.setInputsMapping(inputs_mapping)
    food_model.setOutputsMapping(outputs_mapping)

    # Update optimizer either if we are loading or building a model
    food_model.params = params
    food_model.setOptimizer()
    ###########


    ########### Callbacks
    callbacks = buildCallbacks(params, specific_params, food_model, dataset)
    ###########


    ########### Training
    total_start_time = timer()

    logger.debug('Starting training!')
    training_params = {'n_epochs': params['MAX_EPOCH'], 'batch_size': specific_params['BATCH_SIZE'],
                       'lr_decay': specific_params['LR_DECAY'], 'lr_gamma': specific_params['LR_GAMMA'],
                       'epochs_for_save': params['EPOCHS_FOR_SAVE'], 'verbose': params['VERBOSE'],
                       'n_parallel_loaders': params['PARALLEL_LOADERS'],
                       'extra_callbacks': callbacks, 'reload_epoch': params['RELOAD'], 'epoch_offset': params['RELOAD'],
                       'data_augmentation': params['DATA_AUGMENTATION'],
                       'patience': specific_params['PATIENCE'], 'metric_check': specific_params['STOP_METRIC'],
                       'class_weights': class_weights, 'eval_on_sets': ['test'], 'shuffle': specific_params['SHUFFLE'],
                       }

    food_model.trainNet(dataset, training_params)

    total_end_time = timer()
    time_difference = total_end_time - total_start_time
    logging.info('Total time spent {0:.2f}s = {1:.2f}m'.format(time_difference, time_difference / 60.0))
    ###########
    

def apply_model_to_generate_features(params, specific_params):
    """
        Function for using a previously trained model for sampling.
    """

    ########### Load data
    dataset, inputs_mapping, outputs_mapping, class_weights = build_dataset(params, specific_params, splits=specific_params['SPLITS'])
    ###########

    params = mergeParams(params, specific_params)

    ########### Load model
    model_path = params['STORE_PATH'] % params['MODELS_ROOT_PATH']
    video_model = loadModel(model_path, params['SAMPLING_RELOAD_POINT'], reload_epoch=params['SAMPLING_RELOAD_EPOCH'])
    
    video_model.removeOutputLayers('avg_pool')

    #video_model.setOptimizer()
    ###########


    ########### Apply sampling
    for s in params["EVAL_ON_SETS"]:
        if s == 'test':
            print "...... TEST PREDICTIONS ......"
            paths = dataset.X_test['images']
        elif s == 'train':
            print "...... TRAIN PREDICTIONS ......"
            paths = dataset.X_train['images']

        data = params['INPUTS']['images']

        with open(params['DATA_ROOT_PATH'] + "/" + data['frames_path'] % s, 'r') as f:
            lines = f.readlines()
            frames = [int(l.strip('\n')) for l in lines]

        # Apply model predictions
        params_prediction = {'n_parallel_loaders': params['PARALLEL_LOADERS'],
                             'predict_on_sets': [s]}

        predictions = video_model.predictNet(dataset, params_prediction)[s]

        for i in range(predictions.shape[0]):

            id_feat = paths[i].split('/')[-1][:-4]
            fold = paths[i].split('/')[-2]

            if not os.path.isdir(params['DATA_ROOT_PATH'] + "/features/" + s + '/' + fold):
		        os.mkdir(params['DATA_ROOT_PATH'] + "/features/" + s + '/' + fold)

            # Store result
            filepath = params['DATA_ROOT_PATH'] + '/features/' + s + '/' + fold + '/' + id_feat + '.npy'  # results file
            if params['SAMPLING_SAVE_MODE'] == 'numpy':
                numpy2file(filepath, predictions[i])
            else:
                raise Exception, 'Only "numpy" is allowed in "SAMPLING_SAVE_MODE"'


def buildCallbacks(params, specific_params, model, dataset):
    """
        Builds the selected set of callbacks run during the training of the model
    """

    callbacks = []

    if specific_params['EVALUATE']:

        if 'SORTED_OUTPUTS' in specific_params.keys():
            sorted_keys = specific_params['SORTED_OUTPUTS']
        else:
            sorted_keys = []
            for k in specific_params['OUTPUTS'].keys():
                if specific_params['OUTPUTS'][k]['type'] == 'sigma':
                    sorted_keys.append(k)
                else:
                    sorted_keys.insert(0, k)

        # Evaluate training
        extra_vars = dict()
        extra_vars['n_parallel_loaders'] = params['PARALLEL_LOADERS']

        for i, id_name in enumerate(sorted_keys):
            data = specific_params['OUTPUTS'][id_name]
            if data['type'] != 'sigma':
                extra_vars[i] = dict()


        # Prepare GT for each output
        for s in params['EVAL_ON_SETS']:

            for i, id_name in enumerate(sorted_keys):
                data = specific_params['OUTPUTS'][id_name]

                extra_vars[i]['n_classes'] = specific_params['VOCABULARY_SIZE']

                if data['type'] == 'binary':
                    extra_vars[i][s] = dict()
                    extra_vars[i][s]['word2idx'] = dataset.extra_variables['word2idx_'+id_name]
                elif data['type'] == 'categorical':
                    extra_vars[i][s] = dict()
                elif data['type'] == 'text':
                    extra_vars[i][s] = dict()
                    extra_vars[i][s]['references'] = eval('dataset.Y_' + s + '["' + id_name + '"]')
                    extra_vars[i][s]['words2idx'] = dataset.vocabulary[id_name]['words2idx']
                    extra_vars[i]['words_as_classes'] = data['words_as_classes']
                    extra_vars[i]['multilabel'] = data.get('multilabel', False)
            
                if data['save_classes_report']:
                    extra_vars[i]['save_classes_report'] = data['save_classes_report']
                    with codecs.open(specific_params['DATA_ROOT_PATH']+'/'+data['classes'], 'r', 'utf-8') as f:
                        lines = f.readlines()
                    classes_names = [l.strip('\n') for l in lines]
                    extra_vars[i]['target_names'] = classes_names

        # Prepare other variables for performing evaluation
        gt_id_to_eval = []
        gt_pos_to_eval = []
        vocab = []
        vocab_x = []
        output_types = []
        metrics = []
        write_type = []
        min_pred_val = []

        for i, id_name in enumerate(sorted_keys):
            print id_name
            data = specific_params['OUTPUTS'][id_name]
            if data['type'] == 'binary':
                vocab.append(dataset.extra_variables['idx2word_'+id_name])
            elif data['type'] == 'categorical':
                vocab.append('')
            elif data['type'] == 'sigma':
                break
            elif data['type'] == 'text':
                vocab = dataset.vocabulary[id_name]['idx2words']

            gt_id_to_eval.append(id_name)
            gt_pos_to_eval.append(i)
            vocab_x.append(None)
            output_types.append(data['type'])
            metrics.append(data['metrics'])
            write_type.append(data['write_type'])
            min_pred_val.append(data.get('min_pred_val', 0.5))

        callback_metric = EvalPerformance(model, dataset,
                                           model_name='model',
                                           inputs_mapping_eval=None,
                                           outputs_mapping_eval=None,
                                           gt_id=gt_id_to_eval,
                                           gt_pos=gt_pos_to_eval,
                                           metric_name=metrics,
                                           set_name=params['EVAL_ON_SETS'],
                                           batch_size=specific_params['BATCH_SIZE'],
                                           output_types=output_types,
                                           min_pred_multilabel=min_pred_val,
                                           index2word_y=vocab, # text info
                                           index2word_x=vocab_x,
                                           save_path=model.model_path,
                                           reload_epoch=params['RELOAD'],
                                           start_eval_on_epoch=params['START_EVAL_ON_EPOCH'],
                                           write_samples=params['WRITE_VALID_SAMPLES'],
                                           write_type=write_type,
                                           extra_vars=extra_vars,
                                           verbose=params['VERBOSE'],
                                           do_plot=False)
        callbacks.append(callback_metric)

        callback_lr = LearningRateReducer(reduction_function='step')
        callbacks.append(callback_lr)

    return callbacks


# Merges model-specific params to general params
def mergeParams(params, specific_params):
    for k,v in specific_params.items():
        params[k] = v
    return params


def set_keras_backend(backend):

    if K.backend() != backend:
        os.environ['KERAS_BACKEND'] = backend
        reload(K)
        assert K.backend() == backend


def train_and_evaluate_model_multiple_times(params, specific_params):

    # Reset session (memory) before starting next network training
    if 'tensorflow' == K.backend():
        config = tf.ConfigProto()
        config.gpu_options.per_process_gpu_memory_fraction = params['GPU_MEMORY_FRACTION']
        config.gpu_options.visible_device_list = params['GPU_ID']
        config.gpu_options.allow_growth = True
        set_session(tf.Session(config=config))

    dic_results = {}

    start0 = time.time()
    finished_times = []

    num_evals = params['NUM_EVALS']
    for i in range(num_evals):

        start = time.time()

        ########### Train
        print
        logging.info('======= Start training '+specific_params['MODEL_NAME']+' ======= ' + str(i+1) + "/" + str(num_evals))
        train_model(params, specific_params)
        logging.info('======= Finished training '+specific_params['MODEL_NAME']+' ======= ' + str(i+1) + "/" + str(num_evals))
        print
        ###########

        ########### Evaluate
        print
        logging.info('======= Start evaluating '+specific_params['MODEL_NAME']+' ======= ' + str(i+1) + "/" + str(num_evals))
        if len(specific_params['OUTPUTS_IDS_DATASET']) > 1:
            dic_preds_probs = generate_predictions(params, specific_params)
            dic_preds_probs = combine_predictions_double_output_model(dic_preds_probs, params, specific_params)
            dic_results[i] = apply_nms_and_evaluate_metrics(dic_preds_probs, params, specific_params)
        else:
            dic_preds_probs = generate_predictions(params, specific_params)
            dic_preds_probs = decode_predictions_single_output_model(dic_preds_probs, params, specific_params)
            dic_results[i] = apply_nms_and_evaluate_metrics(dic_preds_probs, params, specific_params)
        logging.info('======= Finished evaluating '+specific_params['MODEL_NAME']+' ======= ' + str(i+1) + "/" + str(num_evals))
        print
        print

        end = time.time()
        finished_times.append(end-start)


    print 
    print "MULTI-EVALUATION RESULTS:"
    print
    print "Num -> Average F1-Score, Weighted Average F1-Score"

    f1_list = []
    w_f1_list = []
    f1_classes_list = []
    for i in range(num_evals):
        print str(i+1) + " -> " + str(dic_results[i]['f1']) + ", " + str(dic_results[i]['w_f1'])
        f1_list.append(dic_results[i]['f1'])
        w_f1_list.append(dic_results[i]['w_f1'])
        f1_classes_list.append(np.append(dic_results[i]['f1_classes'], [0] * (101 - len(dic_results[i]['f1_classes']))))

    mean_f1 = np.mean(f1_list)
    mean_w_f1 = np.mean(w_f1_list)
    median_w_f1 = np.median(w_f1_list)
    idx_median = np.where(w_f1_list == median_w_f1)[0][0]

    with open(params['STORE_PATH'] + "/f1-score_classes_" + specific_params['MODEL_TYPE'] + ".txt", 'w') as f:
        for v in f1_classes_list[idx_median]:
            f.write(str(v)+'\n')
        f.close()

    # Logging in file
    with open(params['STORE_PATH'] + "/multievaluation_" + specific_params['MODEL_TYPE'] + ".txt", "w") as log_f:
        log_f.write(specific_params['MODEL_NAME'] + " MULTI-EVALUATION RESULTS\n")
        log_f.write("\nAverage F1-scores\n")
        for i in range(num_evals): log_f.write(str(dic_results[i]['f1']) + "\n")

        log_f.write("\nWeighted average F1-scores\n")
        for i in range(num_evals): log_f.write(str(dic_results[i]['w_f1']) + "\n")

        log_f.write("\nMean of 'average F1-score': " + str(mean_f1))
        log_f.write("\nMean of 'weighted average F1-score': " + str(mean_w_f1))
        log_f.write("\nMedian of 'weighted average F1-score': " + str(median_w_f1))
        log_f.write("\nCorresponding 'average F1-score' of the weighted median: " + str(f1_list[idx_median]))
        log_f.close()


    print "\nMean of 'average F1-score': " + str(np.mean(f1_list))
    print "Std deviation of 'average F1-score': " + str(np.std(f1_list)) + "\n"

    print "\nMean of 'weighted average F1-score': " + str(np.mean(w_f1_list))
    print "Std deviation of 'weighted average F1-score': " + str(np.std(w_f1_list)) + "\n"

    print "\nMedian of 'weighted average F1-score': " + str(median_w_f1)
    print "Corresponding 'average F1-score' of the weighted median: " + str(f1_list[idx_median]) + "\n"
    print

    end0 = time.time()

    for i, t in enumerate(finished_times):
        print "Computation " + str(i+1) + " lasted " + str(int(t/60)) + " minutes."

    t = end0 - start0
    print
    print "Total time: " + str(int(t/3600)) + " hours and " + str(int((t%3600)/60)) + " minutes.\n"


if __name__ == "__main__":

    # Use 'config_file' command line parameter for changing the name of the config file used
    cf = 'config_dataset'
    for arg in sys.argv[1:]:
        k, v = arg.split('=')
        if k == 'config_file':
            cf = v
    cf = __import__(cf)
    all_models_params = cf.load_parameters()  # parameters for the current dataset version
    params = load_general_parameters()  # general parameters

    # Load keras, mkw and training_wrapper modules.
    sys.path.insert(0, params["MKW_PATH"])
    sys.path.insert(0, params["KERAS_PATH"])

    from keras_wrapper.cnn_model import saveModel, loadModel
    from keras_wrapper.extra import evaluation
    from keras_wrapper.extra.callbacks import EvalPerformance, LearningRateReducer  
    from keras_wrapper.extra.read_write import *
    from keras_wrapper.utils import decode_multilabel, decode_predictions_words_as_classes

    from training_wrapper.prepare_data import build_dataset
    from training_wrapper.model import Food_Model

    from keras import backend as K
    from keras.callbacks import LearningRateScheduler 

    # Configure GPU memory usage for TF
    if 'tensorflow' == K.backend():
        import tensorflow as tf
        from keras.backend.tensorflow_backend import set_session

    if params['MODE'] == 'training':
        logging.info('Running training.')

        model_name = params['MODEL']
        specific_params = all_models_params['ALL_MODELS'][model_name]

        # Reset session (memory) before starting next network training
        if 'tensorflow' == K.backend():
            config = tf.ConfigProto()
            config.gpu_options.per_process_gpu_memory_fraction = params['GPU_MEMORY_FRACTION']
            config.gpu_options.visible_device_list = params['GPU_ID']
            config.gpu_options.allow_growth = True
            set_session(tf.Session(config=config))

        specific_params['STORE_PATH'] = specific_params['STORE_PATH'] % specific_params['MODELS_ROOT_PATH']
        print ("")
        logging.info('======= Start training '+specific_params['MODEL_NAME']+' =======')
        train_model(params, specific_params)
        logging.info('======= Finished training '+specific_params['MODEL_NAME']+' =======')
        print ("")


    elif params['MODE'] == 'predict':
        logging.info('Running sampling.')
        all_models_params = cf.load_parameters()
        
        model_name = params['MODEL']
        specific_params = all_models_params['ALL_MODELS'][model_name]

        if params['YOLO_ENSEMBLE']:
            specific_params['STORE_PATH'] = specific_params['STORE_PATH'] % specific_params['MODELS_ROOT_PATH'] + "_ENS_YOLO"
        else:
            specific_params['STORE_PATH'] = specific_params['STORE_PATH'] % specific_params['MODELS_ROOT_PATH']

        if len(specific_params['OUTPUTS_IDS_DATASET']) > 1:
            dic_preds_probs = generate_predictions(params, specific_params)
            dic_preds_probs = combine_predictions_double_output_model(dic_preds_probs, params, specific_params)
            apply_nms_and_predict(dic_preds_probs, params, specific_params)
            #apply_nms_and_evaluate_metrics(dic_preds_probs, params, specific_params)
        else:
            dic_preds_probs = generate_predictions(params, specific_params)
            dic_preds_probs = decode_predictions_single_output_model(dic_preds_probs, params, specific_params)
            apply_nms_and_predict(dic_preds_probs, params, specific_params)

        logging.info('Done predicting!')


    elif params['MODE'] == 'evaluate':
        logging.info('Running model evaluation.')
        all_models_params = cf.load_parameters()

        for model_name in params['MODELS_TO_EVAL']:
            specific_params = all_models_params['ALL_MODELS'][model_name]

            if params['YOLO_ENSEMBLE']:
                specific_params['STORE_PATH'] = specific_params['STORE_PATH'] % specific_params['MODELS_ROOT_PATH'] + "_ENS_YOLO"
            else:
                specific_params['STORE_PATH'] = specific_params['STORE_PATH'] % specific_params['MODELS_ROOT_PATH']
            
            train_and_evaluate_model_multiple_times(params, specific_params)

        logging.info('Done multi evaluation!')


    elif params['MODE'] == 'features':
        logging.info('Running feature vectors generation.')
        all_models_params = cf.load_parameters()
        
        specific_params = all_models_params['ALL_MODELS']['features']

        apply_model_to_generate_features(params, specific_params)

        logging.info('Done predicting!')
