from keras.engine import Input
from keras.layers import *
from keras.layers.embeddings import Embedding
from keras.layers.recurrent import LSTM
from keras.layers.advanced_activations import PReLU
from keras.layers.normalization import BatchNormalization
from keras.models import model_from_json, Sequential, Model
from keras.optimizers import Adam
from keras.regularizers import l2
from keras.constraints import unit_norm
from keras import backend as K

from keras.applications.resnet50 import ResNet50
from keras.applications.vgg16 import VGG16
from keras.applications.inception_v3 import InceptionV3
from keras.applications.inception_resnet_v2 import InceptionResNetV2

from keras_wrapper.cnn_model import Model_Wrapper
from keras_wrapper.extra.regularize import Regularize

from .custom_layer import ConcatenateOutputWithSigma
#from keras.layers import ConcatenateOutputWithSigma

import numpy as np
import pickle as pk
#import cPickle as pk
import os
import logging
import shutil
import time
import copy


class Food_Model(Model_Wrapper):

    def __init__(self, params, type='InceptionResNetV2_PlusFC', verbose=1, structure_path=None, weights_path=None,
                 model_name=None, store_path=None,
                 empty_label=False):
        """
            Food_Model object constructor.

            :param params: all hyperparameters of the model.
            :param type: network name type (corresponds to any method defined in the section 'MODELS' of this class). Only valid if 'structure_path' == None.
            :param verbose: set to 0 if you don't want the model to output informative messages
            :param structure_path: path to a Keras' model json file. If we speficy this parameter then 'type' will be only an informative parameter.
            :param weights_path: path to the pre-trained weights file (if None, then it will be randomly initialized)
            :param model_name: optional name given to the network (if None, then it will be assigned to current time as its name)
            :param store_path: path to the folder where the temporal model packups will be stored
            :params empty_label: definies if we are introducing an "_empty_" label for the binary outputs
        """
        super(self.__class__, self).__init__(model_name=model_name,
                                             silence=verbose == 0, models_path=store_path, inheritance=True)

        self.__toprint = ['_model_type', 'name', 'model_path', 'verbose']

        self.verbose = verbose
        self._model_type = type
        self.params = params

        # Sets the model name and prepares the folders for storing the models
        self.setName(model_name, store_path)

        # Get sorted list of outputs
        if 'SORTED_OUTPUTS' in params.keys():
            sorted_keys = params['SORTED_OUTPUTS']
        else:
            sorted_keys = []
            for k in params['OUTPUTS'].keys():
                if params['OUTPUTS'][k]['type'] == 'sigma':
                    sorted_keys.append(k)
                else:
                    sorted_keys.insert(0, k)

        # Prepare lists of labels and store them for decoding predictions
        self.labels_list = {}
        for id_out in sorted_keys:
            data = self.params['OUTPUTS'][id_out]
            self.labels_list[id_out] = {'labels': []}
            self.labels_list[id_out]['type'] = data['type']
            self.labels_list[id_out]['min_pred_val'] = data.get('min_pred_val', 0.5)
            if data['type'] != 'sigma': # only ready GT labels if output is not of sigma type
                with open(self.params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as fc:
                    for line in fc:
                        self.labels_list[id_out]['labels'].append(line.rstrip('\n'))
                    if data['type'] == 'binary' and empty_label:
                        self.labels_list[id_out]['labels'].append('_empty_')
            logging.info(self.labels_list[id_out]['labels'])

        # Prepare model
        if structure_path:
            # Load a .json model
            if self.verbose > 0:
                logging.info("<<< Loading model structure from file "+ structure_path +" >>>")
            self.model = model_from_json(open(structure_path).read())
        else:
            # Build model from scratch
            if hasattr(self, type):
                if self.verbose > 0:
                    logging.info("<<< Building "+ type +" Food_Model >>>")
                eval('self.'+type+'(params)')
            else:
                raise Exception('Food_Model type "'+ type +'" is not implemented.')

        # Load weights from file
        if weights_path:
            if self.verbose > 0:
                logging.info("<<< Loading weights from file "+ weights_path +" >>>")
            self.model.load_weights(weights_path, by_name=True, skip_mismatch=True)

        # Print information of self
        """if verbose > 0:
            print (str(self))
            self.model.summary()"""

        self.setOptimizer()


    def setOptimizer(self, metrics=['acc']):

        """
            Sets a new optimizer for the model.
        """
        uncertainty_loss = self.params.get('USE_UNCERTAINTY_LAYER', False)
        if 'SORTED_OUTPUTS' in self.params.keys():
            sorted_keys = self.params['SORTED_OUTPUTS']
        else:
            sorted_keys = []
            for k in self.params['OUTPUTS'].keys():
                if self.params['OUTPUTS'][k]['type'] == 'sigma':
                    sorted_keys.append(k)
                else:
                    sorted_keys.insert(0, k)
        loss = [self.params['OUTPUTS'][k]['loss'] for k in sorted_keys]

        def recall_m(y_true, y_pred):
            true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
            possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
            recall = true_positives / (possible_positives + K.epsilon())
            return recall

        def precision_m(y_true, y_pred):
            true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
            predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
            precision = true_positives / (predicted_positives + K.epsilon())
            return precision

        def f1_m(y_true, y_pred):
            precision = precision_m(y_true, y_pred)
            recall = recall_m(y_true, y_pred)
            return 2*((precision*recall)/(precision+recall+K.epsilon()))

        super(self.__class__, self).setOptimizer(lr=self.params['LR'],
                                                 loss=loss if not uncertainty_loss else None,
                                                 optimizer=self.params['OPTIMIZER'],
                                                 momentum=self.params['MOMENTUM'],
                                                 loss_weights=self.params.get('LOSS_WIGHTS', None),
                                                 sample_weight_mode='temporal' if self.params.get('SAMPLE_WEIGHTS', False) else None,
                                                 metrics=[f1_m],
                                                 epsilon=self.params['EPSILON'],
                                                 tf_optimizer=False)


    def setName(self, model_name, store_path=None, clear_dirs=True):
        """
            Changes the name (identifier) of the model instance.
        """
        if model_name is None:
            self.name = time.strftime("%Y-%m-%d") + '_' + time.strftime("%X")
            create_dirs = False
        else:
            self.name = model_name
            create_dirs = True

        if store_path is None:
            self.model_path = 'Models/' + self.name
        else:
            self.model_path = store_path


        # Remove directories if existed
        if clear_dirs:
            if os.path.isdir(self.model_path):
                shutil.rmtree(self.model_path)

        # Create new ones
        if create_dirs:
            if not os.path.isdir(self.model_path):
                os.makedirs(self.model_path)

    def predict_and_decode(self, X, dataset=None, sort_outs=True):
        '''
            Applies a forward pass on the samples loaded in 'X' and applies
            the decoding of all the outputs provided by the network.

            Returns list with an element per image, where each element is a
            dictionary containing the following information:
                    {
                'label1': {
                                        'labels': ['string1', 'string2', ..., 'stringN'],
                                        'probs': [prob1, prob2, ..., probN],
                                    },
                        'label2': {
                                        'labels': ['string1', 'string2', ..., 'stringM'],
                                        'probs': [prob1, prob2, ..., probM],
                                    },
                    }
        '''

        # Predict output
        preds = self.predictOnBatch(X)

        if 'SORTED_OUTPUTS' in self.params.keys():
            sorted_keys = self.params['SORTED_OUTPUTS']
        else:
            sorted_keys = []
            for k in self.params['OUTPUTS'].keys():
                if self.params['OUTPUTS'][k]['type'] == 'sigma':
                    sorted_keys.append(k)
                else:
                    sorted_keys.insert(0, k)

        n_outputs = len(sorted_keys)
        if n_outputs == 1:
            n_samples = len(preds)
        else:
            n_samples = preds[0].shape[0]

        # Decode
        decodification = [{} for im in range(n_samples)]
        for k, id in enumerate(sorted_keys):
            # for k,(id,pos) in enumerate(self.outputsMapping.iteritems()):
            pos = self.outputsMapping[id]
            type = self.labels_list[id]['type']
            if type != 'sigma':
                labels = self.labels_list[id]['labels']
                min_pred_val = self.labels_list[id]['min_pred_val']

                for im in range(n_samples):  # decodify each image
                    if n_outputs == 1:
                        these_preds = preds[im]
                    else:
                        these_preds = preds[pos][im]

                    if k == 0:
                        decodification[im] = {}
                    decodification[im][id] = {'labels': [], 'probs': []}

                    if type == 'categorical':
                        for i, prob in enumerate(these_preds):
                            decodification[im][id]['labels'].append(labels[i])
                            decodification[im][id]['probs'].append(float(prob))
                    elif type == 'binary':
                        for i, prob in enumerate(these_preds):
                            if prob >= min_pred_val:
                                decodification[im][id]['labels'].append(labels[i])
                                decodification[im][id]['probs'].append(float(prob))
                    else:
                        raise NotImplementedError("Decodification for type '" + type + "' is not implemented.")

                    if sort_outs:
                        order = list(np.argsort(decodification[im][id]['probs']))[::-1]
                        decodification[im][id]['labels'] = [decodification[im][id]['labels'][o] for o in order]
                        decodification[im][id]['probs'] = [decodification[im][id]['probs'][o] for o in order]
        return decodification


    def removeOutputLayers(self, layer_name):

        # Recover last layer before the final dense one
        x = self.model.get_layer(layer_name).output

        self.model = Model(input=self.model.input, output=x)


    # ------------------------------------------------------- #
    #       VISUALIZATION
    #           Methods for visualization
    # ------------------------------------------------------- #

    def __str__(self):
        """
            Plot basic model information.
        """
        obj_str = '-----------------------------------------------------------------------------------\n'
        class_name = self.__class__.__name__
        obj_str += '\t\t'+class_name +' instance\n'
        obj_str += '-----------------------------------------------------------------------------------\n'

        # Print pickled attributes
        for att in self.__toprint:
            obj_str += att + ': ' + str(self.__dict__[att])
            obj_str += '\n'

        obj_str += '\n'
        obj_str += 'MODEL PARAMETERS:\n'
        obj_str += str(self.params)
        obj_str += '\n'

        obj_str += '-----------------------------------------------------------------------------------'

        return obj_str


    # ------------------------------------------------------- #
    #       PREDEFINED MODELS
    # ------------------------------------------------------- #

    def ResNet50_PlusFC(self, params):

        assert len(params['INPUTS'].keys()) == 1, 'Number of inputs must be one.'
        assert params['INPUTS'][params['INPUTS'].keys()[0]]['type'] == 'raw-image', 'Input must be of type "raw-image".'

        self.ids_inputs = params['INPUTS'].keys()
        self.ids_outputs = params['OUTPUTS'].keys()

        input_shape = params['INPUTS'][params['INPUTS'].keys()[0]]['img_size_crop']
        image = Input(name=self.ids_inputs[0], shape=input_shape)

        ##################################################
        # Load Inception model pre-trained on ImageNet
        self.model = ResNet50(weights='imagenet', input_tensor=image)

        # Recover input layer
        #image = self.model.get_layer(self.ids_inputs[0]).output

        # Recover last layer kept from original model: 'fc2'
        x = self.model.get_layer('avg_pool').output
        ##################################################
        x = Flatten()(x)
        # Define outputs
        outputs_list = []
        for id_name, data in params['OUTPUTS'].iteritems():

            # Count the number of output classes
            num_classes = 0
            with open(params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as f:
                for line in f: num_classes += 1

            # Define only a FC output layer (+ activation) per output
            out = Dense(num_classes)(x)
            out_act = Activation(data['activation'], name=id_name)(out)
            outputs_list.append(out_act)


        self.model = Model(input=image, output=outputs_list)


    def InceptionV3_PlusFC(self, params):

        assert len(params['INPUTS'].keys()) == 1, 'Number of inputs must be one.'
        assert params['INPUTS'][params['INPUTS'].keys()[0]]['type'] == 'raw-image', 'Input must be of type "raw-image".'

        self.ids_inputs = params['INPUTS'].keys()
        self.ids_outputs = params['OUTPUTS'].keys()

        input_shape = params['INPUTS'][params['INPUTS'].keys()[0]]['img_size_crop']
        image = Input(name=self.ids_inputs[0], shape=input_shape)

        ##################################################
        # Load Inception model pre-trained on ImageNet
        self.model = InceptionV3(weights='imagenet', input_tensor=image)

        # Recover input layer
        #image = self.model.get_layer(self.ids_inputs[0]).output

        # Recover last layer kept from original model: 'fc2'
        x = self.model.get_layer('avg_pool').output
        ##################################################
        #x = Flatten()(x)
        # Define outputs
        outputs_list = []
        for id_name, data in params['OUTPUTS'].iteritems():

            # Count the number of output classes
            num_classes = 0
            with open(params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as f:
                for line in f: num_classes += 1

            # Define only a FC output layer (+ activation) per output
            out = Dense(num_classes)(x)
            out_act = Activation(data['activation'], name=id_name)(out)
            outputs_list.append(out_act)


        self.model = Model(input=image, output=outputs_list)


    def InceptionResNetV2_Multitask(self, params):

        assert len(params['INPUTS'].keys()) == 1, 'Number of inputs must be one.'
        assert params['INPUTS'][params['INPUTS'].keys()[0]]['type'] == 'raw-image', 'Input must be of type "raw-image".'

        self.ids_inputs = params['INPUTS'].keys()
        self.ids_outputs = params['OUTPUTS'].keys()

        input_shape = params['INPUTS'][params['INPUTS'].keys()[0]]['img_size_crop']
        image = Input(name=self.ids_inputs[0], shape=input_shape)

        ##################################################
        # Load Inception model pre-trained on ImageNet
        self.model = InceptionResNetV2(weights='imagenet', input_tensor=image)

        # Recover input layer
        #image = self.model.get_layer(self.ids_inputs[0]).output

        # Recover last layer kept from original model: 'fc2'
        x = self.model.get_layer('avg_pool').output
        ##################################################
        #x = Flatten()(x)
        # Define outputs
        outputs_list = []
        outputs_matching = {}
        num_classes_matching = {}

        if 'SORTED_OUTPUTS' in params.keys():
            sorted_keys = params['SORTED_OUTPUTS']
        else:
            sorted_keys = []
            for k in params['OUTPUTS'].keys():
                if params['OUTPUTS'][k]['type'] == 'sigma':
                    sorted_keys.append(k)
                else:
                    sorted_keys.insert(0, k)

        #for id_name, data in params['OUTPUTS'].iteritems():
        for id_name in sorted_keys:
            data = params['OUTPUTS'][id_name]

            # Special output that calculates sigmas for uncertainty loss
            if data['type'] == 'sigma':
                match_output = params['OUTPUTS'][id_name]['output_id']
                match_act = outputs_matching[match_output]

                out_sigma = ConcatenateOutputWithSigma((None, num_classes_matching[match_output]+1), name_suffix=id_name, name=id_name)(match_act)
                outputs_list.append(out_sigma)

            else:
                # Count the number of output classes
                num_classes = 0
                with open(params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as f:
                    for line in f: num_classes += 1
                if data['type'] == 'binary' and empty_label:
                    num_classes += 1 # empty label

                # Define only a FC output layer (+ activation) per output
                out = Dense(num_classes)(x)
                out_act = Activation(data['activation'], name=id_name)(out)
                outputs_list.append(out_act)

                outputs_matching[id_name] = out_act
                num_classes_matching[id_name] = num_classes


        self.model = Model(input=image, output=outputs_list)


    def InceptionResNetV2_PlusFC(self, params):

        assert len(list(params['INPUTS'].keys())) == 1, 'Number of inputs must be one.'

        self.ids_inputs = list(params['INPUTS'].keys())
        self.ids_outputs = list(params['OUTPUTS'].keys())

        if params['INPUTS'][list(params['INPUTS'].keys())[0]]['type'] == 'raw-image':
            input_shape = params['INPUTS'][list(params['INPUTS'].keys())[0]]['img_size_crop']
            image = Input(name=self.ids_inputs[0], shape=input_shape)

            ##################################################
            # Load Inception model pre-trained on ImageNet
            self.model = InceptionResNetV2(weights='imagenet', input_tensor=image)

            # Recover last layer kept from original model: 'fc2'
            x = self.model.get_layer('avg_pool').output
            ##################################################

            # Define outputs
            outputs_list = []
            for id_name, data in params['OUTPUTS'].items():

                # Create dropout layer if specified in parameters
                drop_rate = params.get('DROPOUT', None)
                if drop_rate is not None:
                    x = Dropout(drop_rate)(x)

                # Define only a FC output layer (+ activation) per output
                out = Dense(params['VOCABULARY_SIZE'])(x)
                out_act = Activation(data['activation'], name=id_name)(out)
                outputs_list.append(out_act)

            self.model = Model(input=image, output=outputs_list)
            #self.model.summary()


        elif params['INPUTS'][list(params['INPUTS'].keys())[0]]['type'] == 'video':
        
            input_video_shape = tuple([None] + params['INPUTS'][list(params['INPUTS'].keys())[0]]['img_size_crop'])
            input = Input(name=self.ids_inputs[0], shape=input_video_shape)


            ##################################################
            # Load Inception model pre-trained on ImageNet
            core_input = Input(name="core_model_input", shape=params['INPUTS'][list(params['INPUTS'].keys())[0]]['img_size_crop'])
            core_model = InceptionResNetV2(weights='imagenet', input_tensor=core_input)

            # Recover last layer kept from original model: 'fc2'
            x = core_model.get_layer('avg_pool').output 

            core_model = Model(inputs=core_input, outputs=x)
            #core_model.summary()
            ##################################################

            self.model = Sequential()

            self.model.add(
                TimeDistributed(
                    core_model,
                    input_shape=input_video_shape
                )
            )

            # Define outputs        
            data = params['OUTPUTS']['output']

            # Create dropout layer if specified in parameters
            drop_rate = params.get('DROPOUT', None)
            if drop_rate is not None:
                x = TimeDistributed(Dropout(drop_rate))(x)

            # Count the number of output classes
            num_classes = 0
            with open(params['DATA_ROOT_PATH']+'/'+data['classes'], 'r') as f:
                for line in f: num_classes += 1
            if params['EMPTY_LABEL']:
                num_classes += 1
            if data['type'] == 'text':
                num_classes += 3  # Add classes for extra words in vocabulary

            self.model.add(
                TimeDistributed(
                    Dense(num_classes, activation=data['activation'])
                )
            )

            self.model.summary()


    # SINGLE OUTPUT MODELS

    def FV_model(self, params):

        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'

        self.ids_inputs = list(params['INPUTS'].keys())
        self.ids_outputs = list(params['OUTPUTS'].keys())

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################

        # Batch Normalization + ReLu Activation
        input_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_2')(features_input)
        input_features = TimeDistributed(Activation('relu'), name='features_td_relu_2')(input_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_features = TimeDistributed(Dropout(drop_rate), name='features_td_dropout_2')(input_features)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output0"),
                                         name=self.ids_outputs[ind])

        softout1 = shared_FC_soft_1(input_features)

        self.model = Model(input=features_input, output=softout1)

        self.model.summary()


    def FV_C_model(self, params):

        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys())  #['features', 'yolo_predictions', 'correlations', 'menu'] 
        self.ids_outputs = list(params['OUTPUTS'].keys()) #['output', 'output2']  

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        input_fusion_correlations = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_correlations], 
                                 mode='sum', name='fusion_layer')
        out_fusion_layer = Activation('tanh')(out_fusion_layer)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        # Batch Normalization + ReLu Activation
        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################

        self.model = Model(inputs=[features_input, correlations_input], outputs=[softout1])

        self.model.summary()


    def FV_YP_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys())  #['features', 'yolo_predictions', 'correlations', 'menu'] 
        self.ids_outputs = list(params['OUTPUTS'].keys()) #['output', 'output2']  

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################
    

        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################

        
        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions], 
                                 mode='sum', name='fusion_layer')
        out_fusion_layer = Activation('tanh')(out_fusion_layer)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        # Batch Normalization + ReLu Activation
        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################

        self.model = Model(inputs=[features_input, yolo_predictions_input], outputs=[softout1])

        self.model.summary()


    def FV_M_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 1 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys())  #['features', 'yolo_predictions', 'correlations', 'menu'] 
        self.ids_outputs = list(params['OUTPUTS'].keys()) #['output', 'output2']  

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        input_fusion_menu = RepeatVector(max_frame, name='menu_repeat_vector_1')(input_fusion_menu)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_menu], 
                                 mode='sum', name='fusion_layer')
        out_fusion_layer = Activation('tanh')(out_fusion_layer)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        # Batch Normalization + ReLu Activation
        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################

        self.model = Model(inputs=[features_input, menu_input], outputs=[softout1])

        self.model.summary()


    def FV_YP_C_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys())  #['features', 'yolo_predictions', 'correlations', 'menu'] 
        self.ids_outputs = list(params['OUTPUTS'].keys()) #['output', 'output2']  

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################
    

        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        input_fusion_correlations = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)
        ##################################################
        

        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions, input_fusion_correlations], 
                                 mode='sum', name='fusion_layer')
        out_fusion_layer = Activation('tanh')(out_fusion_layer)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        # Batch Normalization + ReLu Activation
        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################

        self.model = Model(inputs=[features_input, yolo_predictions_input, correlations_input], outputs=[softout1])

        self.model.summary()


    def FV_C_M_model(self, params):

        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys())  #['features', 'yolo_predictions', 'correlations', 'menu'] 
        self.ids_outputs = list(params['OUTPUTS'].keys()) #['output', 'output2']  

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        input_fusion_correlations = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        input_fusion_menu = RepeatVector(max_frame, name='menu_repeat_vector_1')(input_fusion_menu)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_correlations, input_fusion_menu], 
                                 mode='sum', name='fusion_layer')
        out_fusion_layer = Activation('tanh')(out_fusion_layer)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        # Batch Normalization + ReLu Activation
        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, correlations_input, features_input], outputs=[softout1])
        self.model.summary()


    def FV_YP_C_M_model(self, params):

        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys())  #['features', 'yolo_predictions', 'correlations', 'menu'] 
        self.ids_outputs = list(params['OUTPUTS'].keys()) #['output', 'output2']  

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        input_fusion_correlations = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        input_fusion_menu = RepeatVector(max_frame, name='menu_repeat_vector_1')(input_fusion_menu)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions, input_fusion_correlations, input_fusion_menu], 
                                 mode='sum', name='fusion_layer')
        out_fusion_layer = Activation('tanh')(out_fusion_layer)

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        # Batch Normalization + ReLu Activation
        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1])
        self.model.summary()


    # MULTIPLE OUTPUTS MODELS

    def FV_YP_C_fusion_M_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        input_fusion_correlations_repeated = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions, input_fusion_correlations_repeated], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        input_fusion_menu_repeated = RepeatVector(max_frame, name='menu_repeat_vector_1')(input_fusion_menu)

        final_fusion = merge([out_fusion_layer, input_fusion_menu_repeated], 
                        mode='sum', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()


    def FV_YP_fusion_C_M_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C + M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='sum', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='sum', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()

    
    # Best fusion model
    def FV_YP_mult_C_M_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C + M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='sum', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='mul', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()

    
    def FV_YP_mult_C_M_mult_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C * M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='mul', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='mul', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()


    def FV_YP_concat_C_M_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C + M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='sum', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='concat', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()


    def FV_YP_fusion_C_M_mult_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C * M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='mul', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='sum', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()


    def FV_YP_C_fusion_C_M_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        input_fusion_correlations_repeated = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions, input_fusion_correlations_repeated], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C * M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='sum', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='sum', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()


    def FV_YP_C_fusion_C_M_mult_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        input_fusion_correlations_repeated = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions, input_fusion_correlations_repeated], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C * M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='mul', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='sum', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()


    def FV_YP_C_mult_C_M_mult_model(self, params):
        assert params['INPUTS']['features']['type'] == 'video-features', 'Input 0 must be of type "video-features".'
        assert params['INPUTS']['correlations']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['yolo_predictions']['type'] == 'text', 'Input 1 must be of type "text".'
        assert params['INPUTS']['menu']['type'] == 'text', 'Input 2 must be of type "text".'

        self.ids_inputs = list(params['INPUTS'].keys()) 
        self.ids_outputs = list(params['OUTPUTS'].keys()) 

        max_frame = params['MAX_FRAME']

        ##################################################
        # FEATURES INPUT

        input_data_0 = params['INPUTS']['features']
        ind = self.ids_inputs.index('features')

        input_shape = tuple([max_frame, input_data_0['feat_len']])
        features_input = Input(name='features', shape=input_shape)

        # Batch Normalization + ReLu Activation
        input_fusion_features = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='features_td_batch_norm_1')(features_input)
        input_fusion_features = TimeDistributed(Activation('relu'), name='features_td_relu_1')(input_fusion_features)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_features = TimeDistributed(Dropout(drop_rate))(input_fusion_features)

        input_fusion_features = TimeDistributed(Dense(params['FUSION_INPUT_LEN'], name='features_td_dense_1', W_regularizer=l2(params['WEIGHT_DECAY'])))(input_fusion_features)
        input_fusion_features = Regularize(input_fusion_features, params, name='features_regularize_1')
        ##################################################


        ##################################################
        # YOLO PREDICTIONS INPUT

        input_data_1 = params['INPUTS']['yolo_predictions']
        ind = self.ids_inputs.index('yolo_predictions')

        yolo_predictions_input = Input(name='yolo_predictions', batch_shape=tuple([None, max_frame]), dtype='int32')

        emb_layer = Embedding(params['VOCABULARY_SIZE'],
                               params['EMBEDDING_SIZE'],
                               name='text_embedding',
                               W_regularizer=l2(params['WEIGHT_DECAY']),
                               mask_zero=False)
        emb = emb_layer(yolo_predictions_input)
        input_fusion_yolo_predictions = Regularize(emb, params, name='yolo_preds_regularize_1')

        input_fusion_yolo_predictions = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='yolo_preds_td_batch_norm_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = TimeDistributed(Activation('relu'), name='yolo_preds_td_relu_1')(input_fusion_yolo_predictions)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_yolo_predictions = TimeDistributed(Dropout(drop_rate), name='yolo_preds_td_dropout_1')(input_fusion_yolo_predictions)

        input_fusion_yolo_predictions = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='yolo_preds_td_dense_1')(input_fusion_yolo_predictions)
        input_fusion_yolo_predictions = Regularize(input_fusion_yolo_predictions, params, name='yolo_preds_regularize_2')
        ##################################################


        ##################################################
        # CORRELATIONS INPUT

        input_data_2 = params['INPUTS']['correlations']
        ind = self.ids_inputs.index('correlations')

        correlations_input = Input(name='correlations', batch_shape=tuple([None, input_data_2['max_text_len']]), dtype='int32')

        emb = emb_layer(correlations_input)
        input_fusion_correlations = Regularize(emb, params, name='correlations_regularize_1')

        sum_dim1 = Lambda(lambda xin: K.mean(xin, axis=1))
        input_fusion_correlations = sum_dim1(input_fusion_correlations)

        input_fusion_correlations = BatchNormalization(epsilon=params['EPSILON'], name='correlations_batch_norm_1')(input_fusion_correlations)
        input_fusion_correlations = Activation('relu', name='correlations_relu_1')(input_fusion_correlations)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_correlations = Dropout(drop_rate, name='correlations_dropout_1')(input_fusion_correlations)

        input_fusion_correlations = Dense(params['FUSION_INPUT_LEN'], name='correlations_dense_1')(input_fusion_correlations)
        input_fusion_correlations = Regularize(input_fusion_correlations, params, name='correlations_regularize_2')
        ##################################################


        ##################################################
        # MENU INPUT

        input_data_3 = params['INPUTS']['menu']
        ind = self.ids_inputs.index('menu')

        menu_input = Input(name='menu', batch_shape=tuple([None, input_data_3['max_text_len']]), dtype='int32')

        emb = emb_layer(menu_input)
        input_fusion_menu = Regularize(emb, params, name='menu_regularize_1')
        input_fusion_menu = sum_dim1(input_fusion_menu)

        input_fusion_menu = BatchNormalization(epsilon=params['EPSILON'], name='menu_batch_norm_1')(input_fusion_menu)
        input_fusion_menu = Activation('relu', name='menu_relu_1')(input_fusion_menu)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            input_fusion_menu = Dropout(drop_rate, name='menu_dropout_1')(input_fusion_menu)

        input_fusion_menu = Dense(params['FUSION_INPUT_LEN'], name='menu_dense_1')(input_fusion_menu)
        input_fusion_menu = Regularize(input_fusion_menu, params, name='menu_regularize_2')
        ##################################################


        ##################################################
        # FUSION + OUTPUT 1 (with background and repetitions)

        input_fusion_correlations_repeated = RepeatVector(max_frame, name='correlations_repeat_vector_1')(input_fusion_correlations)

        out_fusion_layer = merge([input_fusion_features, input_fusion_yolo_predictions, input_fusion_correlations_repeated], 
                                 mode='sum', name='FV_YP_fusion')

        output_data_1 = params['OUTPUTS']['output']
        ind = self.ids_outputs.index('output')

        out_fusion_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='FV_YP_fusion_td_batch_norm_1')(out_fusion_layer)
        out_fusion_layer = TimeDistributed(Activation('relu'), name='FV_YP_fusion_td_relu_1')(out_fusion_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out_fusion_layer = TimeDistributed(Dropout(drop_rate), name='FV_YP_fusion_td_dropout_1')(out_fusion_layer)

        # Output layer: Softmax
        shared_FC_soft_1 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_1"),
                                         name=self.ids_outputs[ind])
        softout1 = shared_FC_soft_1(out_fusion_layer)
        ##################################################


        ##################################################
        # FUSION C * M

        fusion_2 = merge([input_fusion_correlations, input_fusion_menu], 
                                 mode='mul', name='C_M_fusion')

        fusion_2 = BatchNormalization(epsilon=params['EPSILON'], name='C_M_fusion_batch_norm_1')(fusion_2)
        fusion_2 = Activation('relu', name='C_M_fusion_relu_1')(fusion_2)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            fusion_2 = Dropout(drop_rate, name='C_M_fusion_dropout_1')(fusion_2)

        fusion_2 = Dense(params['FUSION_INPUT_LEN'], name='C_M_fusion_dense_1')(fusion_2)
        fusion_2 = Regularize(fusion_2, params, name='C_M_fusion_regularize_1')
        fusion_2 = RepeatVector(max_frame, name='C_M_fusion_repeat_vector_1')(fusion_2)
        ##################################################


        ##################################################
        # FUSION + OUTPUT 2 (with background and repetitions)

        out_fusion_layer = TimeDistributed(Dense(params['FUSION_INPUT_LEN']), name='FV_YP_fusion_dense_1')(out_fusion_layer)
        out_fusion_layer = Regularize(out_fusion_layer, params, name='FV_YP_fusion_regularize_1')

        final_fusion = merge([out_fusion_layer, fusion_2], 
                        mode='mul', name='final_fusion')

        output_data_2 = params['OUTPUTS']['output2']
        ind = self.ids_outputs.index('output2')

        out2_layer = TimeDistributed(BatchNormalization(epsilon=params['EPSILON']), name='final_fusion_td_batch_norm_1')(final_fusion)
        out2_layer = TimeDistributed(Activation('relu'), name='final_fusion_td_relu_1')(out2_layer)

        # Create dropout layer if specified in parameters
        drop_rate = params.get('DROPOUT', None)
        if drop_rate is not None:
            out2_layer = TimeDistributed(Dropout(drop_rate), name='final_fusion_td_dropout_1')(out2_layer)

        # Output layer: Softmax
        shared_FC_soft_2 = TimeDistributed(Dense(params['VOCABULARY_SIZE'],
                                               activation=output_data_1['activation'],
                                               W_regularizer=l2(params['WEIGHT_DECAY']),
                                               name="output_2"),
                                         name=self.ids_outputs[ind])
        softout2 = shared_FC_soft_2(out2_layer)
        ##################################################

        self.model = Model(inputs=[menu_input, yolo_predictions_input, correlations_input, features_input], outputs=[softout1, softout2])
        self.model.summary()
